# -*- coding: utf-8 -*-
#
# Copyright (c) 2015-2017 Kevin Deldycke <kevin@deldycke.com>
#                         and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

""" Expose package-wide elements. """


# Examples of valid version strings
# __version__ = '1.2.3.dev1'  # Development release 1
# __version__ = '1.2.3a1'     # Alpha Release 1
# __version__ = '1.2.3b1'     # Beta Release 1
# __version__ = '1.2.3rc1'    # RC Release 1
# __version__ = '1.2.3'       # Final Release
# __version__ = '1.2.3.post1' # Post Release 1
__version__ = '1.5.4'


import sys

PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3


# Defines custom exception first to avoid circular imports.

class ForbiddenCoordinates(Exception):

    """ A position given as 2D coordinates are out of board's bounds. """


class ForbiddenIndex(Exception):

    """ A position given as an index is out of board's bounds. """


class OccupiedPosition(Exception):

    """ A piece is added to a position already occupied by another. """


class VulnerablePosition(Exception):

    """ A piece is added to a position reachable by another. """


class AttackablePiece(Exception):

    """ A piece is added to a position from which it can attack another. """


# Expose important classes to the root of the module. These are not
# lexicographically sorted to avoid cyclic imports.
from chessboard.pieces import (
    Piece,
    King, Queen, Rook, Bishop, Knight,
    PIECE_LABELS, PIECE_CLASSES)
from chessboard.board import Board
from chessboard.solver import Permutations, SolverContext
from chessboard.benchmark import Benchmark

import logging

logger = logging.getLogger(__name__)



# -*- coding: utf-8 -*-

from __future__ import (
    absolute_import,
    division,
    print_function,
    unicode_literals
)

import os
import subprocess
import sys
import time


# Fetch general information about the project.
# Source: https://github.com/jaraco/skeleton/blob/skeleton/docs/conf.py
root = os.path.join(os.path.dirname(__file__), '..')
setup_script = os.path.join(root, 'setup.py')
fields = ['--name', '--version', '--url', '--author']
dist_info_cmd = [sys.executable, setup_script] + fields
output_bytes = subprocess.check_output(dist_info_cmd, cwd=root)
project_id, version, url, author = output_bytes.decode(
    'utf-8').strip().split('\n')

# Title-case each word of the project ID.
project = ' '.join([word.title() for word in project_id.split('-')])
htmlhelp_basename = project_id

release = version

# Addons.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.extlinks',
    'sphinx.ext.todo',
    'sphinx.ext.viewcode']

master_doc = 'index'

# We use our own copyright template instead of the default as the latter strip
# HTML content.
html_show_copyright = False
copyright = (
    "2015-{}, <a href='https://kevin.deldycke.com'>{}</a> and <a href='{}"
    "/graphs/contributors'>contributors</a>").format(
        time.strftime('%Y'), author, url)

exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

nitpicky = True

# We need a recent sphinx because of the last update format.
needs_sphinx = '1.4'
html_last_updated_fmt = 'YYYY-MM-dd'
templates_path = ['templates']

# Keep the same ordering as in original source code.
autodoc_member_order = 'bysource'

extlinks = {
    'issue': ('{}/issues/%s'.format(url), '#'),
    'pull': ('{}/pull/%s'.format(url), 'PR #')}

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# Use RTD theme both locally and online. Source: https://github.com/snide
# /sphinx_rtd_theme#using-this-theme-locally-then-building-on-read-the-docs
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import sphinx_rtd_theme
    html_theme = 'sphinx_rtd_theme'
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]



# -*- coding: utf-8 -*-
#
# Copyright (c) 2015-2017 Kevin Deldycke <kevin@deldycke.com>
#                         and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

""" Definition of chess pieces and their behavioural properties. """

from __future__ import (
    absolute_import,
    division,
    print_function,
    unicode_literals
)

from collections import OrderedDict
from itertools import chain
from operator import attrgetter

from chessboard import ForbiddenCoordinates

from . import PY2

if PY2:
    from itertools import izip_longest
else:
    from itertools import zip_longest as izip_longest


class Piece(object):
    """ A generic piece.

    x: horizontal position of the piece.
    y: vertical position of the piece.
    """
    # Simple ASCII string identifying the kind of piece.
    label = None

    # Single unicode character used to represent the piece on a board.
    symbol = None

    # Integer uniquely identifying the type/kind of the piece. Used as a
    # shortcut to the class itself. Also serves as a ranking weight of the
    # territory coverage (see :issue:`5`).
    uid = None

    # Cache territory occupied by pieces at a given position for a fixed board.
    territory_cache = {}

    def __init__(self, board, index):
        """ Place the piece on a board at the provided linear position. """
        self.board = board
        self.index = index
        self._x, self._y = None, None

    def __repr__(self):
        """ Display all relevant object internals. """
        return (
            '<{}: uid={}; label={}, symbol={}; x={}, y={}; index={}>'.format(
                self.__class__.__name__,
                self.uid, self.label, self.symbol,
                self.x, self.y, self.index))

    def compute_coordinates(self):
        """ Compute 2D coordinates of the piece. """
        self._x, self._y = self.board.index_to_coordinates(self.index)

    @property
    def x(self):
        """ Return the piece's horizontal position.

        Property is used here so we only compute position once when needed.
        """
        if self._x is None:
            self.compute_coordinates()
        return self._x

    @property
    def y(self):
        """ Return the piece's vertical position.

        Property is used here so we only compute position once when needed.
        """
        if self._y is None:
            self.compute_coordinates()
        return self._y

    @property
    def bottom_distance(self):
        """ Number of squares separating the piece from board's bottom edge.
        """
        return self.board.height - 1 - self.y

    @property
    def right_distance(self):
        """ Number of squares separating the piece from board's right edge. """
        return self.board.length - 1 - self.x

    @property
    def top_distance(self):
        """ Number of squares separating the piece from board's top edge. """
        return self.y

    @property
    def left_distance(self):
        """ Number of squares separating the piece from board's left edge. """
        return self.x

    @property
    def horizontals(self):
        """ All horizontal squares from the piece's point of view.

        Returns a list of relative movements up to the board's bound.
        """
        horizontal_shifts = set(izip_longest(map(
            lambda i: i - self.x, range(self.board.length)), [], fillvalue=0))
        horizontal_shifts.discard((0, 0))
        return horizontal_shifts

    @property
    def verticals(self):
        """ All vertical squares from the piece's point of view.

        Returns a list of relative movements up to the board's bound.
        """
        vertical_shifts = set(izip_longest([], map(
            lambda i: i - self.y, range(self.board.height)), fillvalue=0))
        vertical_shifts.discard((0, 0))
        return vertical_shifts

    @property
    def diagonals(self):
        """ All diagonal squares from the piece's point of view.

        Returns a list of relative movements up to the board's bound.
        """
        left_top_shifts = map(lambda i: (-(i + 1), -(i + 1)), range(min(
            self.left_distance, self.top_distance)))
        left_bottom_shifts = map(lambda i: (-(i + 1), +(i + 1)), range(min(
            self.left_distance, self.bottom_distance)))
        right_top_shifts = map(lambda i: (+(i + 1), -(i + 1)), range(min(
            self.right_distance, self.top_distance)))
        right_bottom_shifts = map(lambda i: (+(i + 1), +(i + 1)), range(min(
            self.right_distance, self.bottom_distance)))
        return set(chain(
            left_top_shifts, left_bottom_shifts,
            right_top_shifts, right_bottom_shifts))

    @property
    def movements(self):
        """ Return list of relative movements allowed. """
        raise NotImplementedError

    @property
    def territory(self):
        """ Return the cached territory occupied by the piece. """
        cache_key = (
            self.board.length, self.board.height, self.uid, self.index)
        if cache_key not in self.territory_cache:
            vector = self.compute_territory()
            self.territory_cache[cache_key] = vector
        else:
            vector = self.territory_cache[cache_key]
        return vector

    def compute_territory(self):
        """ Compute territory reachable by the piece from its current position.

        Returns a list of boolean flags of squares indexed linearly, for which
        a True means the square is reachable.
        """
        # Initialize the square occupancy vector of the board.
        vector = self.board.new_vector()

        # Mark current position as reachable.
        vector[self.index] = True

        # List all places reacheable by the piece from its current position.
        for x_shift, y_shift in self.movements:
            # Mark side positions as reachable if in the limit of the board.
            try:
                reachable_index = self.board.coordinates_to_index(
                    self.x, self.y, x_shift, y_shift)
            except ForbiddenCoordinates:
                continue
            vector[reachable_index] = True

        return vector


class King(Piece):
    """ King model. """

    symbol = '♚'

    @property
    def movements(self):
        """ King moves one square in any direction.

        Don't mind out-of-bounds relative positions: forbidden ones will be
        silently discarded within the ``Piece.territory()`` method above.
        """
        return set([
            # Horizontal movements.
            (+1, 0), (-1, 0),
            # Vertical movements.
            (0, +1), (0, -1),
            # Diagonal movements.
            (+1, +1), (-1, -1), (-1, +1), (+1, -1),
        ])


class Queen(Piece):
    """ Queen model. """

    symbol = '♛'

    @property
    def movements(self):
        """ Queen moves unrestricted horizontally, vertically and diagonally.
        """
        return self.horizontals | self.verticals | self.diagonals


class Rook(Piece):
    """ Rook model. """

    symbol = '♜'

    @property
    def movements(self):
        """ Rook moves unrestricted horizontally and vertically. """
        return self.horizontals | self.verticals


class Bishop(Piece):
    """ Bishop model. """

    symbol = '♝'

    @property
    def movements(self):
        """ Bishop moves unrestricted diagonally. """
        return self.diagonals


class Knight(Piece):
    """ Knight model. """

    symbol = '♞'

    @property
    def movements(self):
        """ Knight moves in L shapes in all 8 directions.

        Don't mind out-of-bounds relative positions: forbidden ones will be
        silently discarded within the ``Piece.territory()`` method above.
        """
        return set([
            # Top-right movements.
            (+2, +1), (+1, +2),
            # Top-left movements.
            (-2, +1), (-1, +2),
            # Bottom-right movements.
            (+2, -1), (+1, -2),
            # Bottom-left movements.
            (-2, -1), (-1, -2),
        ])


# Initialize territory coverage ranking.
Queen.uid, Rook.uid, Bishop.uid, King.uid, Knight.uid = range(5)

# Keep pieces sorted by rank.
PIECES = sorted([Queen, Rook, Bishop, King, Knight], key=attrgetter('uid'))

# Initialize piece labels.
for klass in PIECES:
    klass.label = klass.__name__.lower()

# Map labels to UIDs.
PIECE_LABELS = OrderedDict([(klass.label, klass.uid) for klass in PIECES])

# Map piece UIDs to their class.
PIECE_CLASSES = OrderedDict([(klass.uid, klass) for klass in PIECES])



# -*- coding: utf-8 -*-
#
# Copyright (c) 2015-2017 Kevin Deldycke <kevin@deldycke.com>
#                         and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from __future__ import (
    absolute_import,
    division,
    print_function,
    unicode_literals
)

import unittest

from chessboard import Bishop, Board, King, Knight, Queen, Rook


class TestKing(unittest.TestCase):

    def test_territory(self):
        """ Test computation of territory at each positions of a 3x3 board. """
        board = Board(3, 3)
        c2i = board.coordinates_to_index
        self.assertEquals(King(board, c2i(1, 1)).territory, [
            True, True, True,
            True, True, True,
            True, True, True,
        ])
        self.assertEquals(King(board, c2i(0, 0)).territory, [
            True, True, False,
            True, True, False,
            False, False, False,
        ])
        self.assertEquals(King(board, c2i(1, 0)).territory, [
            True, True, True,
            True, True, True,
            False, False, False,
        ])
        self.assertEquals(King(board, c2i(2, 0)).territory, [
            False, True, True,
            False, True, True,
            False, False, False,
        ])
        self.assertEquals(King(board, c2i(2, 1)).territory, [
            False, True, True,
            False, True, True,
            False, True, True,
        ])
        self.assertEquals(King(board, c2i(2, 2)).territory, [
            False, False, False,
            False, True, True,
            False, True, True,
        ])
        self.assertEquals(King(board, c2i(1, 2)).territory, [
            False, False, False,
            True, True, True,
            True, True, True,
        ])
        self.assertEquals(King(board, c2i(0, 2)).territory, [
            False, False, False,
            True, True, False,
            True, True, False,
        ])
        self.assertEquals(King(board, c2i(0, 1)).territory, [
            True, True, False,
            True, True, False,
            True, True, False,
        ])


class TestQueen(unittest.TestCase):

    def test_territory(self):
        """ Test computation of territory at each positions of a 3x3 board. """
        board = Board(3, 3)
        c2i = board.coordinates_to_index
        self.assertEquals(Queen(board, c2i(1, 1)).territory, [
            True, True, True,
            True, True, True,
            True, True, True,
        ])
        self.assertEquals(Queen(board, c2i(0, 0)).territory, [
            True, True, True,
            True, True, False,
            True, False, True,
        ])
        self.assertEquals(Queen(board, c2i(1, 0)).territory, [
            True, True, True,
            True, True, True,
            False, True, False,
        ])
        self.assertEquals(Queen(board, c2i(2, 0)).territory, [
            True, True, True,
            False, True, True,
            True, False, True,
        ])
        self.assertEquals(Queen(board, c2i(2, 1)).territory, [
            False, True, True,
            True, True, True,
            False, True, True,
        ])
        self.assertEquals(Queen(board, c2i(2, 2)).territory, [
            True, False, True,
            False, True, True,
            True, True, True,
        ])
        self.assertEquals(Queen(board, c2i(1, 2)).territory, [
            False, True, False,
            True, True, True,
            True, True, True,
        ])
        self.assertEquals(Queen(board, c2i(0, 2)).territory, [
            True, False, True,
            True, True, False,
            True, True, True,
        ])
        self.assertEquals(Queen(board, c2i(0, 1)).territory, [
            True, True, False,
            True, True, True,
            True, True, False,
        ])


class TestRook(unittest.TestCase):

    def test_territory(self):
        """ Test computation of territory at each positions of a 3x3 board. """
        board = Board(3, 3)
        c2i = board.coordinates_to_index
        self.assertEquals(Rook(board, c2i(1, 1)).territory, [
            False, True, False,
            True, True, True,
            False, True, False,
        ])
        self.assertEquals(Rook(board, c2i(0, 0)).territory, [
            True, True, True,
            True, False, False,
            True, False, False,
        ])
        self.assertEquals(Rook(board, c2i(1, 0)).territory, [
            True, True, True,
            False, True, False,
            False, True, False,
        ])
        self.assertEquals(Rook(board, c2i(2, 0)).territory, [
            True, True, True,
            False, False, True,
            False, False, True,
        ])
        self.assertEquals(Rook(board, c2i(2, 1)).territory, [
            False, False, True,
            True, True, True,
            False, False, True,
        ])
        self.assertEquals(Rook(board, c2i(2, 2)).territory, [
            False, False, True,
            False, False, True,
            True, True, True,
        ])
        self.assertEquals(Rook(board, c2i(1, 2)).territory, [
            False, True, False,
            False, True, False,
            True, True, True,
        ])
        self.assertEquals(Rook(board, c2i(0, 2)).territory, [
            True, False, False,
            True, False, False,
            True, True, True,
        ])
        self.assertEquals(Rook(board, c2i(0, 1)).territory, [
            True, False, False,
            True, True, True,
            True, False, False,
        ])


class TestBishop(unittest.TestCase):

    def test_territory(self):
        """ Test computation of territory at each positions of a 3x3 board. """
        board = Board(3, 3)
        c2i = board.coordinates_to_index
        self.assertEquals(Bishop(board, c2i(1, 1)).territory, [
            True, False, True,
            False, True, False,
            True, False, True,
        ])
        self.assertEquals(Bishop(board, c2i(0, 0)).territory, [
            True, False, False,
            False, True, False,
            False, False, True,
        ])
        self.assertEquals(Bishop(board, c2i(1, 0)).territory, [
            False, True, False,
            True, False, True,
            False, False, False,
        ])
        self.assertEquals(Bishop(board, c2i(2, 0)).territory, [
            False, False, True,
            False, True, False,
            True, False, False,
        ])
        self.assertEquals(Bishop(board, c2i(2, 1)).territory, [
            False, True, False,
            False, False, True,
            False, True, False,
        ])
        self.assertEquals(Bishop(board, c2i(2, 2)).territory, [
            True, False, False,
            False, True, False,
            False, False, True,
        ])
        self.assertEquals(Bishop(board, c2i(1, 2)).territory, [
            False, False, False,
            True, False, True,
            False, True, False,
        ])
        self.assertEquals(Bishop(board, c2i(0, 2)).territory, [
            False, False, True,
            False, True, False,
            True, False, False,
        ])
        self.assertEquals(Bishop(board, c2i(0, 1)).territory, [
            False, True, False,
            True, False, False,
            False, True, False,
        ])


class TestKnight(unittest.TestCase):

    def test_territory(self):
        """ Test computation of territory at each positions of a 3x3 board. """
        board = Board(3, 3)
        c2i = board.coordinates_to_index
        self.assertEquals(Knight(board, c2i(1, 1)).territory, [
            False, False, False,
            False, True, False,
            False, False, False,
        ])
        self.assertEquals(Knight(board, c2i(0, 0)).territory, [
            True, False, False,
            False, False, True,
            False, True, False,
        ])
        self.assertEquals(Knight(board, c2i(1, 0)).territory, [
            False, True, False,
            False, False, False,
            True, False, True,
        ])
        self.assertEquals(Knight(board, c2i(2, 0)).territory, [
            False, False, True,
            True, False, False,
            False, True, False,
        ])
        self.assertEquals(Knight(board, c2i(2, 1)).territory, [
            True, False, False,
            False, False, True,
            True, False, False,
        ])
        self.assertEquals(Knight(board, c2i(2, 2)).territory, [
            False, True, False,
            True, False, False,
            False, False, True,
        ])
        self.assertEquals(Knight(board, c2i(1, 2)).territory, [
            True, False, True,
            False, False, False,
            False, True, False,
        ])
        self.assertEquals(Knight(board, c2i(0, 2)).territory, [
            False, True, False,
            False, False, True,
            True, False, False,
        ])
        self.assertEquals(Knight(board, c2i(0, 1)).territory, [
            False, False, True,
            True, False, False,
            False, False, True,
        ])



# -*- coding: utf-8 -*-
#
# Copyright (c) 2015-2017 Kevin Deldycke <kevin@deldycke.com>
#                         and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

""" Utilities to search for valid set of positions. """

from __future__ import (
    absolute_import,
    division,
    print_function,
    unicode_literals
)

from itertools import chain

from chessboard import (
    PIECE_LABELS,
    AttackablePiece,
    Board,
    OccupiedPosition,
    VulnerablePosition
)


class Permutations(object):
    """ Produce permutations of pieces iteratively. """

    def __init__(self, pieces, range_size=None):
        # Transform the description of pieces population into a linear vector
        # sorted by UID. Piece symbols are represented by an integer whose
        # weight indicate the priority in the permutation tree, so that pieces
        # covering the widest area are tested first. See :issue:`5`.
        self.pieces = tuple(chain.from_iterable([
            [symbol] * quantity
            for symbol, quantity in sorted(pieces.items())]))

        # Maximal depth of the tree.
        self.depth = len(self.pieces)

        # Range of permutations.
        self.range_size = range_size if range_size else self.depth
        # Because we're using bytearray to store permutations states, we can't
        # keep track of pieces above the 255 position.
        assert self.range_size < 256

        # Keep track of our current progression in search space. This variable
        # always holds the last permutation we returned.
        self.indexes = None

        # Compute the terminal iteration the generator may reach.
        self.terminal_iteration = bytearray([self.range_size - 1] * self.depth)

    def increment(self):
        """ Increment the last permutation we returned to the next. """
        # Increment position from the deepest place of the tree first.
        for index in reversed(range(self.depth)):
            self.indexes[index] += 1
            # We haven't reached the end of board, no need to adjust upper
            # level.
            if self.indexes[index] < self.range_size:
                break
            # We've reached the end of board. Reset current level and increment
            # the upper level.
            self.indexes[index] = 0

        # Now that we incremented our indexes, we need to deduplicate positions
        # shering the same UIDs, by aligning piece's indexes to their parents.
        # This works thanks to the sort performed on self.pieces
        # initialization. See :issue:`7`.
        for i in range(self.depth - 1):
            if (self.pieces[i] == self.pieces[i + 1]) and (
                    self.indexes[i] > self.indexes[i + 1]):
                self.indexes[i + 1] = self.indexes[i]

    def __iter__(self):
        """ Return the generator itself, not its results. """
        return self

    def __next__(self):
        """ Return next valid permutation.

        Raise iteration exception when we explored all permutations.
        """
        if self.indexes is None:
            self.indexes = bytearray(self.depth)
        elif self.indexes == self.terminal_iteration:
            raise StopIteration
        else:
            self.increment()
        return zip(self.pieces, self.indexes)

    next = __next__

    def skip_branch(self, level):
        """ Abandon the branch at the provided level and skip to the next.

        When we call out to skip to the next branch of the search space, we
        push sublevel pieces to the maximum positions of the board. So that the
        next time the permutation iterator is called, it can produce the vector
        state of the next adjacent branch. See :issue:`3`.
        """
        for i in range(level + 1, self.depth):
            self.indexes[i] = self.range_size - 1


class SolverContext(object):
    """ Initialize a chessboard context and search for all possible positions.

    The search space is constrained by board dimensions and piece population.
    """

    def __init__(self, length, height, **pieces):
        """ Initialize board dimensions and piece population."""
        self.length = length
        self.height = height
        assert isinstance(self.length, int)
        assert isinstance(self.height, int)
        assert self.length > 0
        assert self.height > 0

        # Store the number of pieces on the board by their UIDs.
        self.pieces = {}
        for label, quantity in pieces.items():
            assert isinstance(quantity, int)
            assert quantity >= 0
            self.pieces[PIECE_LABELS[label]] = quantity
        assert sum(self.pieces.values()) > 0

        # Solver metadata.
        self.result_counter = 0

    def __repr__(self):
        """ Display all relevant object internals. """
        uids_to_labels = {uid: label for label, uid in PIECE_LABELS.items()}
        return '<SolverContext: length={}, height={}, pieces={}>'.format(
            self.length, self.height, {
                uids_to_labels[uid]: quantity
                for uid, quantity in self.pieces.items()})

    @property
    def vector_size(self):
        return self.length * self.height

    def solve(self):
        """ Solve all possible positions of pieces within the context.

        Depth-first, tree-traversal of the product space.
        """
        # Create a new, empty board.
        board = Board(self.length, self.height)

        # Iterate through all combinations of positions.
        permutations = Permutations(self.pieces, self.vector_size)
        for positions in permutations:

            # Reuse board but flush all pieces.
            board.reset()

            for level, (piece_uid, linear_position) in enumerate(positions):
                # Try to place the piece on the board.
                try:
                    board.add(piece_uid, linear_position)
                # If one of the piece can't be added, throw the whole set, skip
                # the rotten branch and proceed to the next.
                except (OccupiedPosition, VulnerablePosition, AttackablePiece):
                    permutations.skip_branch(level)
                    break

            else:
                # All pieces fits, save solution and proceeed to the next
                # permutation.
                self.result_counter += 1
                yield board



# -*- coding: utf-8 -*-
#
# Copyright (c) 2015-2017 Kevin Deldycke <kevin@deldycke.com>
#                         and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

""" Board repesentation and utilities. """

from __future__ import (
    absolute_import,
    division,
    print_function,
    unicode_literals
)

from operator import or_

from chessboard import (
    PIECE_CLASSES,
    AttackablePiece,
    ForbiddenCoordinates,
    ForbiddenIndex,
    OccupiedPosition,
    VulnerablePosition
)


class Board(object):
    """ Chessboard of arbitrary dimensions with placed pieces.

    This kind of chessboard only accept new pieces which are not overlapping
    squares:

    * occupied by another piece;
    * directly reachable by another piece.

    Internal states of the board are materialized by a vector. A vector is a
    simple iterable for which each element represent a square.

    For :class:`.Piece` we use a :keyword:`bytearray` so we can pack a lot of
    states in memory for caching. But here for boards, we prefer a common
    :keyword:`list` of :keyword:`bool` as it seems Python is a little bit
    faster dealing with these.

    2D positions on the board are noted ``(x, y)``::

          0 1 2 3 4 …
        0 . . . . .
        1 . . . . .
        2 . . . . .
        3 . . . . .
        4 . . . . .
        …

    * horizontal range ``x`` goes from ``0`` to ``m-1``.
    * vertical range ``y`` goes from ``0`` to ``n-1``.
    * top-left position is ``(0, 0)``.
    * top-right position is ``(0, m-1)``.
    * bottom-left position is ``(n-1, 0)``.
    * bottom-right position is ``(n-1, m-1)``.
    """

    def __init__(self, length, height):
        """ Initialize board dimensions. """
        self.length = length
        self.height = height

        # Number of squares in the board.
        self.size = self.length * self.height

        # Ordered list of linear indexes of all squares.
        self.indexes = range(self.size)

        # Call reset() to initialize internal states.
        self.reset()

    def __repr__(self):
        """ Display all relevant object internals. """
        return '<{}: length={}, height={}, size={}, pieces={}>'.format(
            self.__class__.__name__,
            self.length, self.height, self.size, self.pieces)

    def __str__(self):
        """ Render the board with pieces in Unicode-art. """
        lines = []
        # Draw top line.
        lines.append((('┬───' * self.length) + '┐').replace('┬', '┌', 1))
        # Draw each line.
        for y in range(self.height):
            # Draw line with the pieces.
            line = ''
            for x in range(self.length):
                piece = self.get(x, y)
                line += '│ {} '.format(piece.symbol if piece else ' ')
            lines.append(line + '│')
            # Draw line separator but the last one.
            if y < (self.height - 1):
                lines.append(
                    (('┼───' * self.length) + '┤').replace('┼', '├', 1))
        # Draw bottom line.
        lines.append((('┴───' * self.length) + '┘').replace('┴', '└', 1))
        return '\n'.join(lines)

    def reset(self):
        """ Empty board, remove all pieces and reset internal states. """
        # Store positionned pieces on the board.
        self.pieces = set()

        # Squares on the board already occupied by a piece.
        self.occupancy = self.new_vector()

        # Territory susceptible to attacke, i.e. squares reachable by at least
        # a piece.
        self.exposed_territory = self.new_vector()

    @property
    def positions(self):
        """ Generator producing all 2D positions of all squares. """
        for y in range(self.height):
            for x in range(self.length):
                yield x, y

    def new_vector(self):
        """ Returns a list of boolean flags of squares indexed linearly.

        All states are initialized to :keyword:`False`.
        """
        return [False] * self.size

    def validate_index(self, index):
        """ Check that a linear index of a square is within board's bounds. """
        if index < 0 or index >= self.size:
            raise ForbiddenIndex("Linear index {} not in {}x{} board.".format(
                index, self.length, self.height))

    def validate_coordinates(self, x, y):
        """ Check if the piece lie within the board. """
        if not(0 <= x < self.length and 0 <= y < self.height):
            raise ForbiddenCoordinates(
                "x={}, y={} outside of {}x{} board.".format(
                    x, y, self.length, self.height))

    def index_to_coordinates(self, index):
        """ Return a set of 2D ``(x, y)`` coordinates from a linear index. """
        self.validate_index(index)
        x = int(index % self.length)
        y = int((index - x) / self.length)
        return x, y

    def coordinates_to_index(self, x, y, x_shift=0, y_shift=0):
        """ Return a linear index from a set of 2D coordinates.

        Optionnal vertical and horizontal shifts might be applied.
        """
        target_x = x + x_shift
        target_y = y + y_shift
        self.validate_coordinates(target_x, target_y)
        index = (target_y * self.length) + target_x
        return index

    def add(self, piece_uid, index):
        """ Add a piece to the board at the provided linear position. """
        # Square already occupied by another piece.
        if self.occupancy[index]:
            raise OccupiedPosition

        # Square reachable by another piece.
        if self.exposed_territory[index]:
            raise VulnerablePosition

        # Create a new instance of the piece.
        klass = PIECE_CLASSES[piece_uid]
        piece = klass(self, index)

        # Check if a piece can attack another one from its position.
        territory = piece.territory
        for i in self.indexes:
            if self.occupancy[i] and territory[i]:
                raise AttackablePiece

        # Mark the territory covered by the piece as exposed and secure its
        # position on the board.
        self.pieces.add(piece)
        self.occupancy[index] = True
        self.exposed_territory = list(
            map(or_, self.exposed_territory, territory))

    def get(self, x, y):
        """ Return piece placed at the provided coordinates. """
        for piece in self.pieces:
            if (piece.x, piece.y) == (x, y):
                return piece



# -*- coding: utf-8 -*-
#
# Copyright (c) 2015-2017 Kevin Deldycke <kevin@deldycke.com>
#                         and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from __future__ import absolute_import, division, print_function

import logging
import multiprocessing
import time

import click
from click.exceptions import BadParameter

import click_log
from bprofile import BProfile
from chessboard.benchmark import run_scenario

from . import PIECE_LABELS, Benchmark, SolverContext, __version__, logger


class PositiveInt(click.types.IntParamType):
    """ Custom type class for click to validate positive integers. """

    def __init__(self, allow_zero=True):
        """ Validator can be customized to consider 0 as allowed or not. """
        self.allow_zero = allow_zero

    def convert(self, value, param, ctx):
        """ Reuse standard integer validator but add checks on sign and zero.
        """
        value = super(PositiveInt, self).convert(value, param, ctx)
        if value < 0:
            self.fail('%s is not positive' % value, param, ctx)
        if not self.allow_zero and not value:
            self.fail('%s is not greater than 0' % value, param, ctx)
        return value


# Shortcut to pre-configured validators.
POSITIVE_INT = PositiveInt(allow_zero=False)
POSITIVE_OR_ZERO_INT = PositiveInt(allow_zero=True)


class Solve(click.Command):
    """ Manage the solve command. """

    def __init__(self, *args, **kwargs):
        """ Override default constructor to dynamiccaly add pieces options. """
        for label in PIECE_LABELS:
            kwargs['params'].append(click.Option(
                ('--{}'.format(label), ),
                default=0,
                type=POSITIVE_OR_ZERO_INT,
                help='Number of {}s.'.format(label)))
        super(Solve, self).__init__(*args, **kwargs)


@click.group(invoke_without_command=True)
@click_log.init(logger)
@click_log.simple_verbosity_option(
    default='INFO', metavar='LEVEL',
    help='Either CRITICAL, ERROR, WARNING, INFO or DEBUG. Defaults to INFO.')
@click.version_option(__version__)
@click.pass_context
def cli(ctx):
    """ CLI to solve combinatoric chess puzzles. """
    level = click_log.get_level()
    try:
        level_to_name = logging._levelToName
    # Fallback to pre-Python 3.4 internals.
    except AttributeError:
        level_to_name = logging._levelNames
    level_name = level_to_name.get(level, level)
    logger.debug('Verbosity set to {}.'.format(level_name))

    # Print help screen and exit if no sub-commands provided.
    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())
        ctx.exit()


@cli.command(cls=Solve, short_help='Solve a chess puzzle.')
@click.option(
    '-l', '--length', required=True, type=POSITIVE_INT,
    help='Length of the board.')
@click.option(
    '-h', '--height', required=True, type=POSITIVE_INT,
    help='Height of the board.')
@click.option(
    '-s', '--silent', is_flag=True, default=False,
    help='Do not render result boards in ASCII-art.')
@click.option(
    '-p', '--profile', is_flag=True, default=False,
    help='Produce a profiling graph.')
@click.pass_context
def solve(ctx, length, height, silent, profile, **pieces):
    """ Solve a puzzle constrained by board dimensions and pieces. """
    # Check that at least one piece is provided.
    if not sum(pieces.values()):
        context = click.get_current_context()
        raise BadParameter('No piece provided.', ctx=context, param_hint=[
            '--{}'.format(label) for label in PIECE_LABELS])

    # Setup the optionnal profiler.
    profiler = BProfile('solver-profile.png', enabled=profile)

    solver = SolverContext(length, height, **pieces)
    logger.info(repr(solver))

    logger.info('Searching positions...')
    with profiler:
        start = time.time()
        for result in solver.solve():
            if not silent:
                click.echo(u'{}'.format(result))
        processing_time = time.time() - start

    logger.info('{} results found in {:.2f} seconds.'.format(
        solver.result_counter, processing_time))

    if profile:
        logger.info('Execution profile saved at {}'.format(
            profiler.output_path))


@cli.command(short_help='Benchmark the solver.')
def benchmark():
    """ Run a benchmarking suite and measure time taken by the solver.

    Each scenario is run in an isolated process, and results are appended to
    CSV file.
    """
    # Use all cores but one on multi-core CPUs.
    pool_size = multiprocessing.cpu_count() - 1
    if pool_size < 1:
        pool_size = 1

    # Start a pool of workers. Only allow 1 task per child, to force flushing
    # of solver's internal caches.
    pool = multiprocessing.Pool(processes=pool_size, maxtasksperchild=1)
    results = pool.imap_unordered(run_scenario, Benchmark.scenarii)
    pool.close()
    pool.join()

    # Update CSV database with the new results.
    benchmark = Benchmark()
    benchmark.load_csv()
    benchmark.add(results)
    benchmark.save_csv()


@cli.command(short_help='Plot solver performances.')
def graph():
    """ Update all kind of performance graphs from the benchmark data.

    All data come from CSV database.
    """
    benchmark = Benchmark()
    benchmark.load_csv()
    benchmark.nqueen_graph()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2015-2017 Kevin Deldycke <kevin@deldycke.com>
#                         and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from __future__ import (
    absolute_import,
    division,
    print_function,
    unicode_literals
)

import unittest

from chessboard import Board, ForbiddenCoordinates, ForbiddenIndex


class TestBoard(unittest.TestCase):

    def test_all_positions(self):
        self.assertEquals(list(Board(3, 3).positions), [
            (0, 0), (1, 0), (2, 0),
            (0, 1), (1, 1), (2, 1),
            (0, 2), (1, 2), (2, 2)])

    def test_coord_to_index(self):
        self.assertEquals(Board(3, 3).coordinates_to_index(0, 0), 0)
        self.assertEquals(Board(3, 3).coordinates_to_index(1, 1), 4)
        self.assertEquals(Board(3, 3).coordinates_to_index(2, 2), 8)

    def test_translate_error(self):
        with self.assertRaises(ForbiddenCoordinates):
            Board(3, 3).coordinates_to_index(-1, 0)
        with self.assertRaises(ForbiddenCoordinates):
            Board(3, 3).coordinates_to_index(0, -1)
        with self.assertRaises(ForbiddenCoordinates):
            Board(3, 3).coordinates_to_index(0, 3)
        with self.assertRaises(ForbiddenCoordinates):
            Board(3, 3).coordinates_to_index(3, 0)

    def test_index_to_coord(self):
        self.assertEquals(Board(3, 3).index_to_coordinates(0), (0, 0))
        self.assertEquals(Board(3, 3).index_to_coordinates(1), (1, 0))
        self.assertEquals(Board(3, 3).index_to_coordinates(2), (2, 0))
        self.assertEquals(Board(3, 3).index_to_coordinates(3), (0, 1))
        self.assertEquals(Board(3, 3).index_to_coordinates(4), (1, 1))
        self.assertEquals(Board(3, 3).index_to_coordinates(5), (2, 1))
        self.assertEquals(Board(3, 3).index_to_coordinates(6), (0, 2))
        self.assertEquals(Board(3, 3).index_to_coordinates(7), (1, 2))
        self.assertEquals(Board(3, 3).index_to_coordinates(8), (2, 2))

    def test_wide_index_to_coord(self):
        self.assertEquals(Board(1, 4).index_to_coordinates(0), (0, 0))
        self.assertEquals(Board(1, 4).index_to_coordinates(1), (0, 1))
        self.assertEquals(Board(1, 4).index_to_coordinates(2), (0, 2))
        self.assertEquals(Board(1, 4).index_to_coordinates(3), (0, 3))

    def test_long_index_to_coord(self):
        self.assertEquals(Board(4, 1).index_to_coordinates(0), (0, 0))
        self.assertEquals(Board(4, 1).index_to_coordinates(1), (1, 0))
        self.assertEquals(Board(4, 1).index_to_coordinates(2), (2, 0))
        self.assertEquals(Board(4, 1).index_to_coordinates(3), (3, 0))

    def test_index_to_coord_error(self):
        with self.assertRaises(ForbiddenIndex):
            Board(3, 3).index_to_coordinates(-1)
        with self.assertRaises(ForbiddenIndex):
            Board(3, 3).index_to_coordinates(9)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2015-2017 Kevin Deldycke <kevin@deldycke.com>
#                         and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

""" Benchmarking tools. """

from __future__ import (
    absolute_import,
    division,
    print_function,
    unicode_literals
)

import platform
import time
from collections import OrderedDict
from os import path

import pandas
from numpy import median

import seaborn
from chessboard import PIECE_LABELS, SolverContext, __version__
from cpuinfo import get_cpu_info


def run_scenario(params):
    """ Run one scenario and returns execution time and number of solutions.

    Also returns initial parameters in the response to keep the results
    associated with the initial context.
    """
    solver = SolverContext(**params)
    start = time.time()
    count = sum(1 for _ in solver.solve())
    execution_time = time.time() - start
    params.update({
        'solutions': count,
        'execution_time': execution_time})
    return params


class Benchmark(object):

    """ Defines benchmark suite and utility to save and render the results. """

    scenarii = [
        # Tiny boards
        {'length': 3, 'height': 3, 'king': 2, 'rook': 1},
        {'length': 4, 'height': 4, 'rook': 2, 'knight': 4},
        # n queens problems.
        {'length': 1, 'height': 1, 'queen': 1},
        {'length': 2, 'height': 2, 'queen': 2},
        {'length': 3, 'height': 3, 'queen': 3},
        {'length': 4, 'height': 4, 'queen': 4},
        {'length': 5, 'height': 5, 'queen': 5},
        {'length': 6, 'height': 6, 'queen': 6},
        {'length': 7, 'height': 7, 'queen': 7},
        {'length': 8, 'height': 8, 'queen': 8},
        # {'length': 9, 'height': 9, 'queen': 9},
        # Big family.
        {'length': 5, 'height': 5,
         'king': 2, 'queen': 2, 'bishop': 2, 'knight': 1},
        {'length': 6, 'height': 6,
         'king': 2, 'queen': 2, 'bishop': 2, 'knight': 1},
        # {'length': 7, 'height': 7,
        #  'king': 2, 'queen': 2, 'bishop': 2, 'knight': 1},
    ]

    # Data are going in a CSV file along this file.
    csv_filepath = path.join(path.dirname(__file__), 'benchmark.csv')

    # Gather software and hardware metadata.
    cpu_info = get_cpu_info()
    context = OrderedDict([
        # Solver.
        ('chessboard', __version__),
        # Python interpreter.
        ('implementation', platform.python_implementation()),
        ('python', platform.python_version()),
        # Underlaying OS.
        ('system', platform.system()),
        ('macos', platform.mac_ver()[0]),
        ('linux', ' '.join(platform.linux_distribution()).strip()),
        ('windows', platform.win32_ver()[1]),
        ('java', platform.java_ver()[0]),
        # Hardware.
        ('architecture', platform.architecture()[0]),
        ('machine', platform.machine()),
        # CPU.
        ('cpu_vendor', cpu_info['vendor_id']),
        ('cpu_model', cpu_info['brand']),
        ('cpu_freq_actual', cpu_info['hz_actual'][0]),
        ('cpu_freq_advertised', cpu_info['hz_advertised'][0]),
        ('cpu_l2_cache', cpu_info['l2_cache_size']),
    ])

    # Sorted column IDs.
    column_ids = ['length', 'height'] + list(PIECE_LABELS) + [
        'solutions', 'execution_time'] + list(context)

    def __init__(self):
        """ Initialize the result database. """
        self.results = pandas.DataFrame(columns=self.column_ids)

    def load_csv(self):
        """ Load old benchmark results from CSV. """
        if path.exists(self.csv_filepath):
            self.results = self.results.append(
                pandas.read_csv(self.csv_filepath))

    def add(self, new_results):
        """ Add new benchmark results. """
        for result in new_results:
            result.update(self.context)
            self.results = self.results.append(result, ignore_index=True)

    def save_csv(self):
        """ Dump all results to CSV. """
        # Sort results so we can start to see patterns right in the raw CSV.
        self.results.sort_values(by=self.column_ids, inplace=True)
        # Gotcha: integers seems to be promoted to float64 because of
        # reindexation. See: https://pandas.pydata.org/pandas-docs/stable
        # /gotchas.html#na-type-promotions
        self.results.reindex(columns=self.column_ids).to_csv(
            self.csv_filepath, index=False)

    def nqueen_graph(self):
        """ Graph n-queens problem for the current version and context. """
        # Filters out boards with pieces other than queens.
        nqueens = self.results
        for piece_label in set(PIECE_LABELS).difference(['queen']):
            nqueens = nqueens[nqueens[piece_label].map(pandas.isnull)]

        # Filters out non-square boards whose dimension are not aligned to the
        # number of queens.
        nqueens = nqueens[nqueens['length'] == nqueens['queen']]
        nqueens = nqueens[nqueens['height'] == nqueens['queen']]

        # Filters out results not obtained from this system.
        for label, value in self.context.items():
            if not value:
                nqueens = nqueens[nqueens[label].map(pandas.isnull)]
            else:
                nqueens = nqueens[nqueens[label] == value]

        plot = seaborn.factorplot(
            x='queen',
            y='execution_time',
            data=nqueens.sort(columns='queen'),
            estimator=median,
            kind='bar',
            palette='BuGn_d',
            aspect=1.5)
        plot.set_xlabels('Number of queens')
        plot.set_ylabels('Solving time in seconds (log scale)')
        plot.fig.get_axes()[0].set_yscale('log')

        plot.savefig('nqueens-performances.png')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2015-2017 Kevin Deldycke <kevin@deldycke.com>
#                         and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from __future__ import (
    absolute_import,
    division,
    print_function,
    unicode_literals
)

import io
import re
from os import path

from setuptools import find_packages, setup

MODULE_NAME = 'chessboard'
PACKAGE_NAME = MODULE_NAME.replace('_', '-')

DEPENDENCIES = [
    'bprofile',
    'click >= 5.0',
    'click_log',
    'numpy',
    'pandas',
    'py-cpuinfo',
    'seaborn']

EXTRA_DEPENDENCIES = {
    # Extra dependencies are made available through the
    # `$ pip install .[keyword]` command.
    'docs': [
        'sphinx >= 1.4',
        'sphinx_rtd_theme'],
    'tests': [
        'coverage',
        'nose',
        'pipdeptree[graphviz]',
        'pycodestyle >= 2.1.0',
        'pylint'],
    'develop': [
        'bumpversion',
        'isort',
        'readme_renderer >= 16.0',
        'setuptools >= 24.2.1',
        'wheel']}


def read_file(*relative_path_elements):
    """ Return content of a file relative to this ``setup.py``. """
    file_path = path.join(path.dirname(__file__), *relative_path_elements)
    return io.open(file_path, encoding='utf8').read().strip()


# Cache fetched version.
_version = None  # noqa


def version():
    """ Extract version from the ``__init__.py`` file at the module's root.

    Inspired by: https://packaging.python.org/single_source_version/
    """
    global _version
    if _version:
        return _version
    init_file = read_file(MODULE_NAME, '__init__.py')
    matches = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', init_file, re.M)
    if not matches:
        raise RuntimeError("Unable to find version string in __init__.py .")
    _version = matches.group(1)  # noqa
    return _version


def latest_changes():
    """ Extract part of changelog pertaining to version. """
    lines = []
    for line in read_file('CHANGES.rst').splitlines():
        if line.startswith('-------'):
            if len(lines) > 1:
                lines = lines[:-1]
                break
        if lines:
            lines.append(line)
        elif line.startswith("`{} (".format(version())):
            lines.append(line)
    if not lines:
        raise RuntimeError(
            "Unable to find changelog for the {} release.".format(version()))
    # Renormalize and clean lines.
    return '\n'.join(lines).strip().split('\n')


def long_description():
    """ Collates project README and latest changes. """
    changes = latest_changes()
    changes[0] = "`Changes for v{}".format(changes[0][1:])
    changes[1] = '-' * len(changes[0])
    return "\n\n\n".join([
        read_file('README.rst'),
        '\n'.join(changes),
        "`Full changelog <https://{}.readthedocs.io/en/develop/changelog.html"
        "#changelog>`_.".format(PACKAGE_NAME)])


setup(
    name=PACKAGE_NAME,
    version=version(),
    description="CLI to solve combinatoric chess puzzles.",
    long_description=long_description(),
    keywords=['chess', 'puzzle', 'CLI', 'solver'],

    author='Kevin Deldycke',
    author_email='kevin@deldycke.com',
    url='https://github.com/kdeldycke/chessboard',
    license='GPLv2+',

    packages=find_packages(),
    # https://www.python.org/dev/peps/pep-0345/#version-specifiers
    python_requires='>= 2.7, != 3.0.*, != 3.1.*, != 3.2.*',
    install_requires=DEPENDENCIES,
    tests_require=DEPENDENCIES + EXTRA_DEPENDENCIES['tests'],
    extras_require=EXTRA_DEPENDENCIES,
    dependency_links=[],
    test_suite='{}.tests'.format(MODULE_NAME),

    classifiers=[
        # See: https://pypi.python.org/pypi?:action=list_classifiers
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: '
        'GNU General Public License v2 or later (GPLv2+)',
        'Operating System :: OS Independent',
        # List of python versions and their support status:
        # https://en.wikipedia.org/wiki/CPython#Version_history
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Education',
        'Topic :: Games/Entertainment :: Board Games'],

    entry_points={
        'console_scripts': [
            'chessboard={}.cli:cli'.format(MODULE_NAME)]}
)
